#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------- #
# Copyright (c) 2025, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2025. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# ----------------------------------------------------------------------- #
import os
import numpy as np

from aps.beamline_driver.beam_management.control_systems.common import Directions, ScanParameters
from aps.beamline_driver.beam_management.facade import Beam, BeamProperties, MethodCallingType
from aps.beamline_driver.beam_management.abstract_beam_manager import DirectBeamImageBeamManager, DirectBeamImageProperties

from aps.common.plot.image import get_average, get_peak_location, get_fwhm, get_sigma

from epics import PV

from matplotlib.figure import Figure
from matplotlib.ticker import FuncFormatter

import time

class EpicsScanPVs:
    def __init__(self, pv_names : dict):
        self.__counts         = PV(pv_names["counts"])
        self.__acquire        = PV(pv_names["acquire"])
        self.__acquire_time   = PV(pv_names["acquire_time"])
        self.__motor_h        = PV(pv_names["motor_h"])
        self.__motor_v        = PV(pv_names["motor_v"])
        self.__motor_h_moving = PV(pv_names["motor_h_moving"])
        self.__motor_v_moving = PV(pv_names["motor_v_moving"])
        if pv_names["autocount"] is None: self.__autocount = None
        else:                             self.__autocount = PV(pv_names["autocount"])

    @property
    def counts(self): return self.__counts
    @property
    def acquire(self): return self.__acquire
    @property
    def acquire_time(self): return self.__acquire_time
    @property
    def motor_h(self): return self.__motor_h
    @property
    def motor_v(self): return self.__motor_v
    @property
    def motor_h_moving(self): return self.__motor_h_moving
    @property
    def motor_v_moving(self): return self.__motor_v_moving
    @property
    def autocount(self): return self.__autocount

MOTOR_WAIT_TIME = 0.1
ACQUIRE_WAIT_TIME = 0.1

class EpicsScanDirectBeamManager(DirectBeamImageBeamManager):
    def __init__(self, pvs: EpicsScanPVs, **kwargs):
        super(EpicsScanDirectBeamManager, self).__init__(method_calling_type=MethodCallingType.STANDARD, **kwargs)

        self._acquire_value = kwargs.get("acquire_value", "Count")

        self._pvs = pvs

    def get_beam(self, *args, **kwargs) -> Beam:
        if not self._pvs.autocount is None: self._pvs.autocount.put(1)

        direction         = kwargs.get("direction", Directions.BOTH)
        parameters        = kwargs.get("parameters", [ScanParameters(), ScanParameters()])

        beam_data = {}

        def get_profile(coord, scan):
            derivative    = np.gradient(scan, coord)
            # the sign of the derivative depends on how the scan is made,
            # the scan could be descending or ascending,changing the sign of the derivative
            if (abs(np.min(derivative)) > abs(np.max(derivative))): derivative *= -1.0

            normalization = np.max(scan) / np.max(derivative)
            profile       = derivative * normalization

            profile[np.where(profile < 0.0)]     = 0.0
            profile[np.where(np.isnan(profile))] = 0.0

            return profile

        if direction == Directions.HORIZONTAL:
            data = self.__scan(self._pvs.motor_h, self._pvs.motor_h_moving, parameters)
            beam_data["coordinate_h"] = data[:, 0]
            beam_data["profile_h"]    = get_profile(data[:, 0], data[:, 1])
        elif direction == Directions.VERTICAL:
            data = self.__scan(self._pvs.motor_v, self._pvs.motor_v_moving, parameters)
            beam_data["coordinate_v"] = data[:, 0]
            beam_data["profile_v"]    = get_profile(data[:, 0], data[:, 1])
        elif direction == Directions.BOTH:
            self.__move_motor(self._pvs.motor_v, self._pvs.motor_v_moving, parameters[1].shift_position)
            data_h = self.__scan(self._pvs.motor_h, self._pvs.motor_h_moving, parameters[0])

            self.__move_motor(self._pvs.motor_h, self._pvs.motor_h_moving, parameters[0].shift_position)
            data_v = self.__scan(self._pvs.motor_v, self._pvs.motor_v_moving, parameters[1])

            beam_data["coordinate_h"] = data_h[:, 0]
            beam_data["scan_h"]       = data_h[:, 1]
            beam_data["profile_h"]    = get_profile(data_h[:, 0], data_h[:, 1])
            beam_data["coordinate_v"] = data_v[:, 0]
            beam_data["scan_v"]       = data_v[:, 1]
            beam_data["profile_v"]    = get_profile(data_v[:, 0], data_v[:, 1])

        return Beam(beam_data=beam_data)

    def __scan(self, motor, motor_moving, scan_parameters : ScanParameters):
        positions = scan_parameters.coordinates

        data       = np.zeros((scan_parameters.steps, 2), float)
        data[:, 0] = positions

        print(f"set acquire time to {scan_parameters.acquire_time}")
        self._pvs.acquire_time.put(scan_parameters.acquire_time)

        for i in range(scan_parameters.steps):
            print(f"motor {motor.pvname}: {positions[i]}" )
            self.__move_motor(motor, motor_moving, positions[i])
            print("start counting:")
            self._pvs.acquire.put(self._acquire_value)
            time.sleep(ACQUIRE_WAIT_TIME)
            while self._pvs.acquire.get() != 0: time.sleep(0.1)
            data[i, 1] = self._pvs.counts.get()
            print(f"collected {data[i, 1]} counts")
        # put the motor on the peak!
        if scan_parameters.set_motor_on_peak:
            self.__move_motor(motor, motor_moving, data[np.argmax(data[:, 1]), 0])

        return data

    def __move_motor(self, motor, motor_moving, position):
        motor.put(position)

        moving = 0
        while (moving != 1):
            time.sleep(MOTOR_WAIT_TIME)
            moving = motor_moving.get()

    def get_beam_properties(self, beam : Beam, **kwargs) -> BeamProperties:
        beam_data : dict = beam.data

        beam_properties = BeamProperties()

        integral_intensity_h = 0
        peak_intensity_h     = 0
        integral_intensity_v = 0
        peak_intensity_v     = 0

        if "coordinate_h" in beam_data:
            centroid_h            = get_average(beam_data["profile_h"], beam_data["coordinate_h"])
            peak_distance_h       = get_peak_location(beam_data["profile_h"], beam_data["coordinate_h"])
            fwhm_h, _, _          = get_fwhm(beam_data["profile_h"], beam_data["coordinate_h"])
            sigma_h               = get_sigma(beam_data["profile_h"], beam_data["coordinate_h"])
            peak_intensity_h      = np.max(beam_data["profile_h"])
            integral_intensity_h  = np.sum(beam_data["profile_h"])

            beam_properties.set_parameter(DirectBeamImageProperties.CENTROID_H, centroid_h)
            beam_properties.set_parameter(DirectBeamImageProperties.PEAK_LOCATION_H, peak_distance_h)
            beam_properties.set_parameter(DirectBeamImageProperties.FWHM_H, fwhm_h)
            beam_properties.set_parameter(DirectBeamImageProperties.SIGMA_H, sigma_h)
            beam_properties.set_parameter("coordinate_h", beam_data["coordinate_h"])
            beam_properties.set_parameter(self.get_beam_histogram_name() + "_h", beam_data["profile_h"])

        if "coordinate_v" in beam_data:
            centroid_v            = get_average(beam_data["profile_v"], beam_data["coordinate_v"])
            peak_distance_v       = get_peak_location(beam_data["profile_v"], beam_data["coordinate_v"])
            fwhm_v, _, _          = get_fwhm(beam_data["profile_v"], beam_data["coordinate_v"])
            sigma_v               = get_sigma(beam_data["profile_v"], beam_data["coordinate_v"])
            peak_intensity_v      = np.max(beam_data["profile_v"])
            integral_intensity_v  = np.sum(beam_data["profile_v"])

            beam_properties.set_parameter(DirectBeamImageProperties.CENTROID_V, centroid_v)
            beam_properties.set_parameter(DirectBeamImageProperties.PEAK_LOCATION_V, peak_distance_v)
            beam_properties.set_parameter(DirectBeamImageProperties.FWHM_V, fwhm_v)
            beam_properties.set_parameter(DirectBeamImageProperties.SIGMA_V, sigma_v)
            beam_properties.set_parameter("coordinate_v", beam_data["coordinate_v"])
            beam_properties.set_parameter(self.get_beam_histogram_name() + "_v", beam_data["profile_v"])

        peak_intensity     = max(peak_intensity_h, peak_intensity_v)
        integral_intensity = max(integral_intensity_h, integral_intensity_v)

        beam_properties.set_parameter(DirectBeamImageProperties.PEAK_INTENSITY, peak_intensity)
        beam_properties.set_parameter(DirectBeamImageProperties.INTEGRAL_INTENSITY, integral_intensity)

        beam_properties.set_parameter("is_empty_beam", integral_intensity <= self._get_input_parameters().get_parameter("empty_beam_threshold", 0.0))

        self._plot_beam(beam_properties, **kwargs)

        return beam_properties

    def _plot_beam(self, beam_properties, **kwargs):
        if kwargs.get("save_image", False) or kwargs.get("show_beam", False):
            hh     = beam_properties.get_parameter("coordinate_h")
            vv     = beam_properties.get_parameter("coordinate_v")
            data_h = beam_properties.get_parameter(self.get_beam_histogram_name() + "_h")
            data_v = beam_properties.get_parameter(self.get_beam_histogram_name() + "_v")

            xrange = kwargs.get("xrange", None)
            yrange = kwargs.get("yrange", None)

            if xrange is None: xrange = [np.min(hh), np.max(hh)]
            if yrange is None: yrange = [np.min(vv), np.max(vv)]

            # Define a custom formatter function
            def custom_formatter(x, pos): return f'{x:.2f}'

            plt = kwargs.get("plt", None)

            if plt is None: fig = Figure(figsize=[10, 8])
            else:           fig = plt.figure(figsize=[10, 8])

            ax1, ax2, ax3 = fig.subplots(1, 3, gridspec_kw={'width_ratios': [2.5, 2.5, 1.0]}, sharey=True)
            ax3.set_axis_off()

            ax1.set_xlim(xrange[0], xrange[1])
            ax1.set_xticks(np.linspace(xrange[0], xrange[1], 6, endpoint=True))
            ax1.xaxis.set_major_formatter(FuncFormatter(custom_formatter))
            ax1.axvline(0, color="gray", ls="--", linewidth=1, alpha=0.7)
            ax1.set_xlabel("Horizontal ($\\mu$m)")
            ax1.set_ylabel("Intensity")
            ax1.plot(hh, data_h)

            ax2.set_xlim(yrange[0], yrange[1])
            ax2.set_xticks(np.linspace(yrange[0], yrange[1], 6, endpoint=True))
            ax2.xaxis.set_major_formatter(FuncFormatter(custom_formatter))
            ax2.axvline(0, color="gray", ls="--", linewidth=1, alpha=0.7)
            ax2.set_xlabel("Vertical ($\\mu$m)")
            ax2.plot(vv, data_v)

            labels = self.get_labels()

            text = (rf"{'fwhm(H)':<6}  = {beam_properties.get_parameter(labels[0]) : 3.1f} $\mu$m" + "\n" +
                    rf"{'fwhm(V)':<6}  = {beam_properties.get_parameter(labels[1]) : 3.1f} $\mu$m" + "\n" +
                    rf"{'sigma(H)':<6} = {beam_properties.get_parameter(labels[2]) : 3.1f} $\mu$m" + "\n" +
                    rf"{'sigma(V)':<6} = {beam_properties.get_parameter(labels[3]) : 3.1f} $\mu$m" + "\n" +
                    rf"{'peak(H)':<6}  = {beam_properties.get_parameter(labels[4]) : 3.1f} $\mu$m" + "\n" +
                    rf"{'peak(V)':<6}  = {beam_properties.get_parameter(labels[5]) : 3.1f} $\mu$m" + "\n" +
                    rf"{'pI':<10}      = {beam_properties.get_parameter(labels[6]) : 1.3e} ph")

            ax3.text(0.01, 0.7, text, color="black", alpha=0.9, fontsize=9,
                     bbox=dict(facecolor="white", edgecolor="gray", alpha=0.7), transform=ax3.transAxes)

            trial_num = kwargs.get("trial_num", None)
            if not trial_num is None: fig.gca().text(0.02, 0.95, str(trial_num), fontsize=10, color="blue", transform=ax1.transAxes)

            if kwargs.get("save_image", False):
                file_name = kwargs.get("file_name", None)
                if not file_name is None:
                    fig.savefig(os.path.join(kwargs.get("directory_name", os.path.abspath(os.curdir)), file_name))
                    fig.canvas.draw_idle()

            if kwargs.get("show_beam", False):
                plt = kwargs.get("plt", fig)
                plt.show()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------- #
# Copyright (c) 2025, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2025. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# ----------------------------------------------------------------------- #
import os

import numpy as np

from aps.common.plot.image import get_average, get_peak_location, apply_transformations

from matplotlib.figure import Figure
from matplotlib.ticker import FuncFormatter

from aps.beamline_driver.beam_management.facade import Beam, BeamProperties, MethodCallingType
from aps.beamline_driver.beam_management.abstract_beam_manager import WavefrontAnalysisBeamManager, WavefrontAnalysisProperties

from aps.wavefront_analysis.absolute_phase.factory import create_absolute_phase_analyzer
from aps.wavefront_analysis.driver.factory import create_wavefront_sensor


class EpicsWavefrontAnalyzerBeamManager(WavefrontAnalysisBeamManager):
    def __init__(self,
                 measurement_directory=None,
                 energy = 20000,
                 **kwargs):
        super(EpicsWavefrontAnalyzerBeamManager, self).__init__(method_calling_type=MethodCallingType.STANDARD, **kwargs)

        measurement_directory = measurement_directory if measurement_directory is not None else os.path.abspath(os.path.join(os.curdir, "wf_images"))

        self.__wavefront_sensor = create_wavefront_sensor(measurement_directory=measurement_directory,
                                                          file_name_prefix=kwargs.get("file_name_prefix", None),
                                                          detector_delay=kwargs.get("detector_delay", None))
        self.__wavefront_analyzer = create_absolute_phase_analyzer(data_collection_directory=measurement_directory,
                                                                   file_name_prefix=kwargs.get("file_name_prefix", None),
                                                                   energy=energy)

    def _get_beam(self, *args, **kwargs) -> Beam:
        units       = kwargs.get("units", "mm")
        verbose     = kwargs.get("verbose", True)
        save_images = kwargs.get("save_images", False)

        try:    self.__wavefront_sensor.restore_status()
        except: pass

        try:
            self.__wavefront_sensor.collect_single_shot_image(index=1)

            if self.__wavefront_sensor.get_configuration_file().IS_STREAM_AVAILABLE:
                image, _, _ = self.__wavefront_sensor.get_image_stream_data(units=units)
                self.__wavefront_analyzer.process_image(image_index=1, image_data=image, save_images=save_images, verbose=verbose)
            else:
                self.__wavefront_analyzer.process_image(image_index=1, image_data=None, save_images=save_images, verbose=verbose)

            wavefront_data = self.__wavefront_analyzer.back_propagate_wavefront(image_index=1,
                                                                                show_figure=False,
                                                                                save_result=False,
                                                                                verbose=verbose)
            try:    self.__wavefront_sensor.save_status()
            except: pass
            try:    self.__wavefront_sensor.end_collection()
            except: pass
        except Exception as e:
            try:    self.__wavefront_sensor.save_status()
            except: pass
            try:    self.__wavefront_sensor.end_collection()
            except: pass

            raise e

        return Beam(beam_data=wavefront_data)

    def get_hardware_device(self): return self.__wavefront_sensor

    def get_beam_properties(self, beam : Beam, **kwargs) -> BeamProperties:
        wavefront_data  = beam.data
        beam_properties = BeamProperties()

        beam_properties.set_parameter(WavefrontAnalysisProperties.INTENSITY_SIGMA_H, wavefront_data["sigma_x"]*1e6)
        beam_properties.set_parameter(WavefrontAnalysisProperties.INTENSITY_SIGMA_V, wavefront_data["sigma_y"]*1e6)
        beam_properties.set_parameter(WavefrontAnalysisProperties.INTENSITY_FWHM_H, wavefront_data["fwhm_x"]*1e6)
        beam_properties.set_parameter(WavefrontAnalysisProperties.INTENSITY_FWHM_V, wavefront_data["fwhm_y"]*1e6)
        beam_properties.set_parameter(WavefrontAnalysisProperties.INTENSITY_CENTROID_H, wavefront_data["wf_position_x"]*1e6)
        beam_properties.set_parameter(WavefrontAnalysisProperties.INTENSITY_CENTROID_V, wavefront_data["wf_position_y"]*1e6)
        beam_properties.set_parameter(WavefrontAnalysisProperties.FOCUS_Z_POSITION_H, wavefront_data["focus_z_position_x"]*1e3)
        beam_properties.set_parameter(WavefrontAnalysisProperties.FOCUS_Z_POSITION_V, wavefront_data["focus_z_position_y"]*1e3)

        coordinates_h = wavefront_data["coordinates_x"]*1e6
        coordinates_v = wavefront_data["coordinates_y"]*1e6

        beam_properties.set_parameter("coordinate_h", coordinates_h)
        beam_properties.set_parameter("coordinate_v", coordinates_v)

        # -------------------------------------------------------------------------------------------

        if wavefront_data["kind"] == "1D":
            intensity_h = wavefront_data["intensity_x"]
            intensity_v = wavefront_data["intensity_y"]
            beam_properties.set_parameter(WavefrontAnalysisProperties.INTENSITY_MAXIMUM,  np.average(np.max(intensity_h), np.max(intensity_v)))
            beam_properties.set_parameter(WavefrontAnalysisProperties.INTENSITY_INTEGRAL, np.average(np.sum(intensity_h), np.sum(intensity_v)))
        elif wavefront_data["kind"] == "2D":
            intensity_h = wavefront_data["integrated_intensity_x"]
            intensity_v = wavefront_data["integrated_intensity_y"]
            intensity   = wavefront_data["intensity"]
            beam_properties.set_parameter(self.get_beam_histogram_name(), intensity)
            beam_properties.set_parameter(WavefrontAnalysisProperties.INTENSITY_MAXIMUM,  np.max(intensity))
            beam_properties.set_parameter(WavefrontAnalysisProperties.INTENSITY_INTEGRAL, np.sum(intensity))
        else:
            raise ValueError("Kind not recognized")

        beam_properties.set_parameter(self.get_beam_histogram_name() + "_h", intensity_h)
        beam_properties.set_parameter(self.get_beam_histogram_name() + "_v", intensity_v)
        beam_properties.set_parameter("is_empty_beam",
                                      beam_properties.get_parameter(WavefrontAnalysisProperties.INTENSITY_INTEGRAL) <= \
                                      self._get_input_parameters().get_parameter("empty_beam_threshold", 0.0))

        self._plot_beam(beam_properties, **kwargs)

        return beam_properties

    def _plot_beam(self, beam_properties, **kwargs):
        if kwargs.get("save_image", False) or kwargs.get("show_beam", False):
            hh     = beam_properties.get_parameter("coordinate_h")
            vv     = beam_properties.get_parameter("coordinate_v")
            data_h = beam_properties.get_parameter(self.get_beam_histogram_name() + "_h")
            data_v = beam_properties.get_parameter(self.get_beam_histogram_name() + "_v")

            xrange = kwargs.get("xrange", None)
            yrange = kwargs.get("yrange", None)

            if xrange is None: xrange = [np.min(hh), np.max(hh)]
            if yrange is None: yrange = [np.min(vv), np.max(vv)]

            # Define a custom formatter function
            def custom_formatter(x, pos): return f'{x:.2f}'

            plt = kwargs.get("plt", None)

            if plt is None: fig = Figure(figsize=[10, 8])
            else:           fig = plt.figure(figsize=[10, 8])

            ax1, ax2, ax3 = fig.subplots(1, 3, gridspec_kw={'width_ratios': [2.5, 2.5, 1.0]}, sharey=True)
            ax3.set_axis_off()

            ax1.set_xlim(xrange[0], xrange[1])
            ax1.set_xticks(np.linspace(xrange[0], xrange[1], 6, endpoint=True))
            ax1.xaxis.set_major_formatter(FuncFormatter(custom_formatter))
            ax1.axvline(0, color="gray", ls="--", linewidth=1, alpha=0.7)
            ax1.set_xlabel("Horizontal ($\\mu$m)")
            ax1.set_ylabel("Intensity")
            ax1.plot(hh, data_h)

            ax2.set_xlim(yrange[0], yrange[1])
            ax2.set_xticks(np.linspace(yrange[0], yrange[1], 6, endpoint=True))
            ax2.xaxis.set_major_formatter(FuncFormatter(custom_formatter))
            ax2.axvline(0, color="gray", ls="--", linewidth=1, alpha=0.7)
            ax2.set_xlabel("Vertical ($\\mu$m)")
            ax2.plot(vv, data_v)

            labels = self.get_labels()

            text = (rf"{'fwhm(H)':<6}  = {beam_properties.get_parameter(labels[0]) : 3.1f} $\\mu$m" + "\n" +
                    rf"{'fwhm(V)':<6}  = {beam_properties.get_parameter(labels[1]) : 3.1f} $\\mu$m" + "\n" +
                    rf"{'sigma(H)':<6} = {beam_properties.get_parameter(labels[2]) : 3.1f} $\\mu$m" + "\n" +
                    rf"{'sigma(V)':<6} = {beam_properties.get_parameter(labels[3]) : 3.1f} $\\mu$m" + "\n" +
                    rf"{'z pos(H)':<6} = {beam_properties.get_parameter(labels[4]) : 6.3f} mm" + "\n" +
                    rf"{'z pos(V)':<6} = {beam_properties.get_parameter(labels[5]) : 6.3f} mm" + "\n" +
                    rf"{'pI':<10}      = {beam_properties.get_parameter(labels[6]) : 1.3e} ph")

            ax3.text(0.01, 0.7, text, color="black", alpha=0.9, fontsize=9,
                     bbox=dict(facecolor="white", edgecolor="gray", alpha=0.7), transform=ax3.transAxes)

            trial_num = kwargs.get("trial_num", None)
            if not trial_num is None: fig.gca().text(0.02, 0.95, str(trial_num), fontsize=10, color="blue", transform=ax1.transAxes)

            if kwargs.get("save_image", False):
                file_name = kwargs.get("file_name", None)
                if not file_name is None:
                    fig.savefig(os.path.join(kwargs.get("directory_name", os.path.abspath(os.curdir)), file_name))
                    fig.canvas.draw_idle()

            if kwargs.get("show_beam", False):
                plt = kwargs.get("plt", fig)
                plt.show()






#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------- #
# Copyright (c) 2025, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2025. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# ----------------------------------------------------------------------- #
import random

try:    import bluesky.plan_stubs as bps
except: print("Bluesky not installed")

import asyncio
import numpy as np
import abc

from aps.beamline_driver.motion_management.facade import Movement, MotorsInfo, MotorsInfoLegacy, Units, MethodCallingType
from aps.beamline_driver.motion_management.abstract_motion_manager import AbstractMotionManager, Keys

try:
    from shadow4.beamline.s4_beamline_element_movements import S4BeamlineElementMovements
    from shadow4.beamline.s4_beamline_element import S4BeamlineElement
except: print("Shadow4 not installed")

class S4MotionManager(AbstractMotionManager):
    class MotorFunction:
        PITCH  = 0
        ROLL   = 1
        YAW    = 2
        NORMAL = 3
        TANGENTIAL = 4
        SAGITTAL = 5
        BENDER = 6

    class BenderPosition:
        UPSTREAM   = 0
        DOWNSTREAM = 1

    def __init__(self,
                 motors_info: MotorsInfo | MotorsInfoLegacy=None,
                 json_file=None):
        if motors_info is None and json_file is None: motors_info = self.get_default_motors_info()

        super(S4MotionManager, self).__init__(motors_info=motors_info,
                                              json_file=json_file,
                                              method_calling_type=MethodCallingType.STANDARD,
                                              platform=Keys.Platforms.DigitalTwin)

        self._movements_registry = {oe_key : {"N" : [0.0, 0.0], "T" : [0.0, 0.0]} for oe_key in self._get_oe_keys()}

    def set_method_calling_type(self, value):
        super(S4MotionManager, self).set_method_calling_type(value)

        if self.method_calling_type() == MethodCallingType.GENERATOR:
            setattr(self, "_move_motor",         self.__move_motor_generator)
            setattr(self, "_get_motor_position", self.__get_motor_position_generator)
        else:
            setattr(self, "_move_motor",         self.__move_motor)
            setattr(self, "_get_motor_position", self.__get_motor_position)

    @abc.abstractmethod
    def get_default_motors_info(self) -> MotorsInfo | MotorsInfoLegacy: raise NotImplementedError
    @abc.abstractmethod
    def _get_motor_function(self, motor_id) -> int: raise NotImplementedError
    @abc.abstractmethod
    def _get_s4_beamline_element(self, motor_id) -> S4BeamlineElement: raise NotImplementedError
    @abc.abstractmethod
    def _get_bender_position(self, motor_id) -> int: raise NotImplementedError
    @abc.abstractmethod
    def _get_oe_keys(self) -> list: raise NotImplementedError
    @abc.abstractmethod
    def _get_oe_key(self, motor_id) -> str: raise NotImplementedError

    def _move_motor_common(self, motor_id : str, position: float, movement : Movement = Movement.ABSOLUTE, units : Units = None):
        if not movement==Movement.ABSOLUTE: raise ValueError("this motion manager does not supports relative movements")
        if units is not None and units != self.get_motors_info().get_motor_info(motor_id).units:
            raise ValueError("this motion manager does not supports units, it will use default ones")

        motor_info         = self.get_motors_info().get_motor_info(motor_id)
        motor_function     = self._get_motor_function(motor_id)
        s4_optical_element = self._get_s4_beamline_element(motor_id)

        repeatibility      = motor_info.repeatability
        coordinates        = s4_optical_element.get_coordinates()
        movements          = s4_optical_element.get_movements()
        movements_registry = self._movements_registry[self._get_oe_key(motor_id)]
        
        if repeatibility > 0.0: position += random.uniform(-repeatibility, repeatibility)

        if motor_function == S4MotionManager.MotorFunction.PITCH:
            self._set_motor_pitch(pitch=position, coordinates=coordinates, movements=movements, motor_info=motor_info)
        elif motor_function == S4MotionManager.MotorFunction.ROLL:
            self._set_motor_roll(roll=position, movements=movements, motor_info=motor_info)
        elif motor_function == S4MotionManager.MotorFunction.YAW:
            self._set_motor_yaw(yaw=position, movements=movements, motor_info=motor_info)
        elif motor_function == S4MotionManager.MotorFunction.NORMAL:
            self._set_motor_normal(translation=position, coordinates=coordinates, movements=movements, motor_info=motor_info, movements_registry=movements_registry)
        elif motor_function == S4MotionManager.MotorFunction.TANGENTIAL:
            self._set_motor_tangential(translation=position, coordinates=coordinates, movements=movements, motor_info=motor_info, movements_registry=movements_registry)
        elif motor_function == S4MotionManager.MotorFunction.SAGITTAL:
            self._set_motor_sagittal(translation=position, movements=movements, motor_info=motor_info)
        elif motor_function == S4MotionManager.MotorFunction.BENDER:
            bender_movement = s4_optical_element.get_optical_element().bender_movement
            bender_position = self._get_bender_position(motor_id)

            position_upstream   = position if bender_position == S4MotionManager.BenderPosition.UPSTREAM else None
            position_downstream = position if bender_position == S4MotionManager.BenderPosition.DOWNSTREAM else None

            self._set_motor_bender(bender=s4_optical_element,
                                   position_upstream=position_upstream,
                                   position_downstream=position_downstream,
                                   bender_movement=bender_movement,
                                   motor_info=motor_info)

    def _get_motor_position_common(self, motor_id : str, units : Units = None) -> float:
        if units is not None and units != self.get_motors_info().get_motor_info(motor_id).units:
            raise ValueError("this motion manager does not supports units, it will use default ones")

        motor_info         = self.get_motors_info().get_motor_info(motor_id)
        motor_function     = self._get_motor_function(motor_id)
        s4_optical_element = self._get_s4_beamline_element(motor_id)

        coordinates        = s4_optical_element.get_coordinates()
        movements          = s4_optical_element.get_movements()
        movements_registry = self._movements_registry[self._get_oe_key(motor_id)]

        if motor_function == S4MotionManager.MotorFunction.PITCH:
            position = motor_info.apply_resolution(self._get_motor_pitch(coordinates, movements, motor_info))
        elif motor_function == S4MotionManager.MotorFunction.ROLL:
            position = motor_info.apply_resolution(self._get_motor_roll(movements, motor_info))
        elif motor_function == S4MotionManager.MotorFunction.YAW:
            position = motor_info.apply_resolution(self._get_motor_yaw(movements, motor_info))
        elif motor_function == S4MotionManager.MotorFunction.NORMAL:
            position = motor_info.apply_resolution(self._get_motor_normal(coordinates, movements, motor_info, movements_registry))
        elif motor_function == S4MotionManager.MotorFunction.TANGENTIAL:
            position = motor_info.apply_resolution(self._get_motor_tangential(coordinates, movements, motor_info, movements_registry))
        elif motor_function == S4MotionManager.MotorFunction.SAGITTAL:
            position = motor_info.apply_resolution(self._get_motor_sagittal(movements, motor_info))
        elif motor_function == S4MotionManager.MotorFunction.BENDER:
            bender_movement = s4_optical_element.get_optical_element().bender_movement
            bender_position = self._get_bender_position(motor_id)

            if   bender_position == S4MotionManager.BenderPosition.UPSTREAM:   position = motor_info.apply_resolution(bender_movement.position_upstream)
            elif bender_position == S4MotionManager.BenderPosition.DOWNSTREAM: position = motor_info.apply_resolution(bender_movement.position_downstream)
        else:
            raise ValueError("Motor Function not recognized")

        return position

    def __move_motor_generator(self, motor_id : str, position: float, movement : Movement = Movement.ABSOLUTE, units : Units = None):
        self._move_motor_common(motor_id, position, movement, units)
        yield from bps.null() # this call to make this method a bluesky plan

    async def __move_motor_coroutine(self, motor_id : str, position: float, movement : Movement = Movement.ABSOLUTE, units : Units = None):
        self._move_motor_common(motor_id, position, movement, units)
        await asyncio.sleep(0.001)

    def __move_motor(self, motor_id : str, position: float, movement : Movement = Movement.ABSOLUTE, units : Units = None):
        self._move_motor_common(motor_id, position, movement, units)

    def __get_motor_position_generator(self, motor_id : str, units : Units = None) -> float:
        yield from bps.null()
        return self._get_motor_position_common(motor_id, units)

    async def __get_motor_position_coroutine(self, motor_id : str, units : Units = None) -> float:
        await asyncio.sleep(0.001)
        return self._get_motor_position_common(motor_id, units)

    def __get_motor_position(self, motor_id : str, units : Units = None) -> float:
        return self._get_motor_position_common(motor_id, units)

    def _set_motor_bender(self, bender, position_upstream, position_downstream, bender_movement, motor_info):
        if not position_upstream is None:   bender_movement.position_upstream   = motor_info.apply_resolution(position_upstream)
        if not position_downstream is None: bender_movement.position_downstream = motor_info.apply_resolution(position_downstream)

        bender.get_optical_element().move_bender(bender_movement)

    def _set_motor_pitch(self, pitch, coordinates, movements, motor_info):
        movements.rotation_x = self.__to_s4_angle(pitch, motor_info) -  (0.5*np.pi - coordinates.angle_radial())

    def _set_motor_roll(self, roll, movements, motor_info):
        movements.rotation_y = self.__to_s4_angle(roll, motor_info)

    def _set_motor_yaw(self, yaw, movements, motor_info):
        movements.rotation_z = self.__to_s4_angle(yaw, motor_info)

    def _set_motor_normal(self, translation, coordinates, movements, motor_info, movements_registry):
        s4_pitch = self._get_s4_motor_pitch(coordinates, movements) # rad from normal

        z_n = self.__to_s4_length(translation, motor_info) / np.cos(s4_pitch)
        y_n = z_n * np.tan(s4_pitch)

        movements_registry["N"][0] = y_n
        movements_registry["N"][1] = z_n

        movements.offset_y = movements_registry["N"][0] + movements_registry["T"][0]
        movements.offset_z = movements_registry["N"][1] + movements_registry["T"][1]

    def _set_motor_tangential(self, translation, coordinates, movements, motor_info, movements_registry):
        s4_pitch = self._get_s4_motor_pitch(coordinates, movements)

        y_t = self.__to_s4_length(translation, motor_info) / np.cos(s4_pitch)
        z_t = -y_t * np.tan(s4_pitch)

        movements_registry["T"][0] = y_t
        movements_registry["T"][1] = z_t

        movements.offset_y = movements_registry["N"][0] + movements_registry["T"][0]
        movements.offset_z = movements_registry["N"][1] + movements_registry["T"][1]

    def _set_motor_sagittal(self, translation, movements, motor_info):
        movements.offset_x = self.__to_s4_length(translation, motor_info)

    def _get_motor_pitch(self, coordinates, movements, motor_info):
        return self.__from_s4_angle(self._get_s4_motor_pitch(coordinates, movements), motor_info)

    def _get_motor_roll(self, movements, motor_info):
        return self.__from_s4_angle(self._get_s4_motor_roll(movements), motor_info)

    def _get_motor_yaw(self, movements, motor_info):
        return self.__from_s4_angle(self._get_s4_motor_yaw(movements), motor_info)

    def _get_motor_normal(self, coordinates, movements, motor_info, movements_registry):
        return self.__from_s4_length(self._get_s4_motor_normal(coordinates, movements, movements_registry), motor_info)

    def _get_motor_tangential(self, coordinates, movements, motor_info, movements_registry):
        return self.__from_s4_length(self._get_s4_motor_tangential(coordinates, movements, movements_registry), motor_info)

    def _get_motor_sagittal(self, movements, motor_info):
        return self.__from_s4_length(self._get_s4_motor_sagittal(movements), motor_info)

    def _get_s4_motor_pitch(self, coordinates, movements):
        return 0.5*np.pi - coordinates.angle_radial() + movements.get_rotations()[0]

    def _get_s4_motor_roll(self,  movements):
        return movements.get_rotations()[1]

    def _get_s4_motor_yaw(self, movements):
        return movements.get_rotations()[2]

    def _get_s4_motor_normal(self, coordinates, movements, movements_registry):
        s4_pitch = self._get_s4_motor_pitch(coordinates, movements)
        offsets  = movements_registry["N"]

        return offsets[0] * np.sin(s4_pitch) + offsets[1] * np.cos(s4_pitch)

    def _get_s4_motor_tangential(self, coordinates, movements, movements_registry):
        s4_pitch = self._get_s4_motor_pitch(coordinates, movements)
        offsets  = movements_registry["T"]

        return offsets[0] * np.cos(s4_pitch) + offsets[1] * np.sin(s4_pitch)

    def _get_s4_motor_sagittal(self, movements):
        return movements.get_offsets()[0]

    def __to_s4_angle(self, angle, motor_info): # user units to rad
        shift_from_ideal = motor_info.shift_from_ideal

        if   motor_info.units == Units.RADIANS:      return motor_info.apply_resolution(angle - shift_from_ideal)
        elif motor_info.units == Units.MILLIRADIANS: return 1e-3*motor_info.apply_resolution(angle - shift_from_ideal)
        elif motor_info.units == Units.DEGREES:      return np.radians(motor_info.apply_resolution(angle - shift_from_ideal))
        else: raise ValueError("Wrong units on motor " + str(motor_info.legacy_ids))

    def __from_s4_angle(self, s4_angle, motor_info): # rad to user units
        shift_from_ideal = motor_info.shift_from_ideal

        if   motor_info.units == Units.RADIANS:      return s4_angle + shift_from_ideal
        elif motor_info.units == Units.MILLIRADIANS: return 1e3*s4_angle + shift_from_ideal
        elif motor_info.units == Units.DEGREES:      return np.degrees(s4_angle) + shift_from_ideal
        else: raise ValueError("Wrong units on motor " + str(motor_info.legacy_ids))

    def __to_s4_length(self, length, motor_info): # user units to m
        shift_from_ideal = motor_info.shift_from_ideal

        if   motor_info.units == Units.METERS:       return motor_info.apply_resolution(length - shift_from_ideal)
        elif motor_info.units == Units.MILLIMETERS:  return 1e-3*motor_info.apply_resolution(length - shift_from_ideal)
        elif motor_info.units == Units.MICRON:       return 1e-6*motor_info.apply_resolution(length - shift_from_ideal)
        else: raise ValueError("Wrong units on motor " + str(motor_info.legacy_ids))

    def __from_s4_length(self, s4_length, motor_info): # m to user units
        shift_from_ideal = motor_info.shift_from_ideal

        if   motor_info.units == Units.METERS:      return s4_length  + shift_from_ideal
        elif motor_info.units == Units.MILLIMETERS: return 1e3*s4_length  + shift_from_ideal
        elif motor_info.units == Units.MICRON:      return 1e6*s4_length  + shift_from_ideal
        else: raise ValueError("Wrong units on motor " + str(motor_info.legacy_ids))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------- #
# Copyright (c) 2025, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2025. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# ----------------------------------------------------------------------- #
from PIL import Image
import os
import cv2 as cv
import numpy as np
from skimage.filters import threshold_multiotsu

import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm

from aps.beamline_driver.common.utility import calculate_projections_over_noise

factor               = 1e6
PIXEL_SIZE           = 6.5e-7
IMAGE_SIZE_PIXEL_HxV = [2160, 2560]

def load_image(file_path):
    if os.path.exists(file_path): return np.array(np.array(Image.open(file_path))).astype(np.float32).T
    else:                         raise ValueError('Error: wrong data path. No data is loaded:' + file_path)

image_raw = load_image(os.path.join(os.path.abspath(os.curdir), "sample_300ms_00001.tif"))

h_coord = np.linspace(- IMAGE_SIZE_PIXEL_HxV[0] / 2, IMAGE_SIZE_PIXEL_HxV[0] / 2, IMAGE_SIZE_PIXEL_HxV[0]) * PIXEL_SIZE * factor
v_coord = np.linspace(- IMAGE_SIZE_PIXEL_HxV[1] / 2, IMAGE_SIZE_PIXEL_HxV[1] / 2, IMAGE_SIZE_PIXEL_HxV[1]) * PIXEL_SIZE * factor

dx = (h_coord[1] - h_coord[0]) / 2.
dy = (v_coord[1] - v_coord[0]) / 2.
extent = [h_coord[0] - dx, h_coord[-1] + dx, v_coord[0] - dy, v_coord[-1] + dy]


### De-noised image with threshold based on upper left 10x10 region

image_0, histogram_h, histogram_v = calculate_projections_over_noise(histogram       = image_raw,
                                                                     noise_max_pos   = 10,
                                                                     noise_threshold = 1.1)

'''
Simple Otsu thresholding
'''


def get_otsu_thresholding(image):
    image_8bit = cv.convertScaleAbs(image, alpha=(255.0 / 65535.0))
    # Apply Otsu's thresholding
    ret, otsu_threshold = cv.threshold(image_8bit, 0, 255, cv.THRESH_BINARY + cv.THRESH_OTSU)
    return otsu_threshold


def get_thresholded_image(image, threshold):
    image_ = np.multiply(image, threshold)
    histogram_h = image_.sum(axis=1)
    histogram_v = image_.sum(axis=0)
    return image_, histogram_h, histogram_v

def plot_image_and_hist(image, title, h_hist, v_hist, h_coord=h_coord, v_coord=v_coord, xlim=None, ylim=None):
    fig, (ax1, ax2, ax3) = plt.subplots(1, 3, gridspec_kw={'width_ratios': [2, 1, 1]})
    fig.set_size_inches(12, 4)
    fig.gca().set_title(title)

    im = ax1.imshow(image, extent=extent, norm=LogNorm(vmin=0.01, vmax=np.max(image)))
    if xlim is None: xlim=[h_coord[0], h_coord[-1]]
    if ylim is None: ylim=[v_coord[0], v_coord[-1]]

    ax1.set_xlim(xlim[0], xlim[-1])
    ax1.set_ylim(ylim[0], ylim[-1])
    ax1.set_xlabel("Horizontal [um]")
    ax1.set_ylabel("Vertical [um]")
    plt.colorbar(mappable=im, ax=ax1)

    ax2.set_yscale('log')
    ax3.set_yscale('log')
    ax2.set_xlabel("Horizontal [um]")
    ax3.set_xlabel("Vertical [um]")
    ax2.plot(h_coord, h_hist)
    ax3.plot(v_coord, v_hist)
    ax2.set_xlim(xlim[0], xlim[-1])
    ax3.set_xlim(ylim[0], ylim[-1])


image_cv_raw = cv.imread(filename=os.path.join(os.path.abspath(os.curdir), "sample_300ms_00001.tif"), flags=cv.IMREAD_UNCHANGED).T

threshold_otsu = get_otsu_thresholding(image=image_cv_raw)
image_otsu, hist_h_otsu, hist_v_otsu = get_thresholded_image(image=image_cv_raw, threshold=threshold_otsu)

#plot_image_and_hist(image=image_raw.T, title='Image_raw', h_hist=image_raw.sum(axis=1), v_hist=image_raw.sum(axis=0), xlim=[-50, 50], ylim=[-50, 50])
#plot_image_and_hist(image=image_otsu.T, title='Image_Otsu', h_hist=hist_h_otsu, v_hist=hist_v_otsu, xlim=[-50, 50], ylim=[-50, 50])
#plot_image_and_hist(image=image_0.T, title='Image_Fixie', h_hist=histogram_h, v_hist=histogram_v, xlim=[-50, 50], ylim=[-50, 50])


'''
Simple Otsu thresholding
'''

MED_FILTER_SIZE = 31 # 11

image_cv_raw_8bit = cv.convertScaleAbs(image_cv_raw, alpha=(255.0/image_cv_raw.max()))
img_MedianBlur    = cv.medianBlur(image_cv_raw_8bit, MED_FILTER_SIZE)
thresholds        = threshold_multiotsu(img_MedianBlur)

regions = np.digitize(image_cv_raw_8bit, bins=thresholds)
regions[regions == 0] = 0
regions[regions == 1] = 1
regions[regions == 2] = 1

image_multi_otsu, hist_h_multi_otsu, hist_v_multi_otsu = get_thresholded_image(image = image_cv_raw, threshold = regions)
plot_image_and_hist(image  = image_multi_otsu.T,
                    title  = 'Image_multi_otsu',
                    h_hist = image_multi_otsu.sum(axis = 1),
                    v_hist = image_multi_otsu.sum(axis = 0),
                    xlim=[-50, 50], ylim=[-50, 50])
plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------- #
# Copyright (c) 2025, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2025. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# ----------------------------------------------------------------------- #
import traceback

import numpy as np
from ophyd import Component
from ophyd import Device, Kind, Signal
from ophyd_async.core import Device as AsyncDevice
from ophyd.sim import NullStatus

from aps.beamline_driver.beam_management.facade import IBeamManager

class _AbstractDetector(Device):
    def __init__(self,
                 prefix: str = "",
                 name: str = "beam_manager",
                 kind: Kind = Kind.normal,
                 beam_manager: IBeamManager = None,
                 verbose: bool = False):
        super(_AbstractDetector, self).__init__(prefix=prefix, name=name, kind=kind)
        self._beam_manager = beam_manager
        self._verbose      = verbose

    def _set_properties(self, beam_string, beam_properties):
        attributes = {}

        for beam_property in self._beam_manager.get_beam_properties_type().to_list():
            value = beam_properties.get_parameter(beam_property)
            getattr(self, beam_property.replace("-", "_")).put(value)
            attributes[beam_property] = value
        getattr(self, "beam").put(beam_string)
        attributes["beam"] = beam_string

        return attributes

    def _set_error(self, error_message):
        attributes = {}
        for beam_property in self._beam_manager.get_beam_properties_type().to_list():
            getattr(self, beam_property.replace("-", "_")).put(np.NaN)
            attributes[beam_property] = np.NaN

        if self._verbose: print("Error while retrieving beam properties: ", error_message)

        return attributes

    def trigger(self, *args, **kwargs):
        try:
            beam            = self._beam_manager.get_beam()
            beam_properties = self._beam_manager.get_beam_properties(beam)

            if beam_properties.get_parameter("is_empty_beam"): attributes = self._set_error("Beam is Empty")
            else:                                              attributes = self._set_properties(beam_string=beam.serialize(),
                                                                                                 beam_properties=beam_properties)
        except Exception:
            attributes = self._set_error(traceback.format_exc())

        if self._verbose: print("called Beam manager!", attributes)

        super().trigger(*args, **kwargs)

        return NullStatus()


class DirectBeamImageDetector(_AbstractDetector):
    centroid_h         = Component(Signal, kind="normal")
    centroid_v         = Component(Signal, kind="normal")
    peak_distance_h    = Component(Signal, kind="normal")
    peak_distance_v    = Component(Signal, kind="normal")
    fwhm_h             = Component(Signal, kind="normal")
    fwhm_v             = Component(Signal, kind="normal")
    sigma_h            = Component(Signal, kind="normal")
    sigma_v            = Component(Signal, kind="normal")
    peak_intensity     = Component(Signal, kind="normal")
    integral_intensity = Component(Signal, kind="normal")
    beam               = Component(Signal, kind="normal")

    def __init__(self,
                 prefix: str = "",
                 name: str = "beam_manager",
                 kind: Kind = Kind.normal,
                 beam_manager: IBeamManager = None,
                 verbose: bool = False):
        super(DirectBeamImageDetector, self).__init__(prefix=prefix,
                                                      name=name,
                                                      kind=kind,
                                                      beam_manager=beam_manager,
                                                      verbose=verbose)

class WavefrontAnalysisDetector(_AbstractDetector):
    intensity_centroid_h             = Component(Signal, kind="normal")
    intensity_centroid_v             = Component(Signal, kind="normal")
    intensity_peak_distance_h        = Component(Signal, kind="normal")
    intensity_peak_distance_v        = Component(Signal, kind="normal")
    intensity_fwhm_h                 = Component(Signal, kind="normal")
    intensity_fwhm_v                 = Component(Signal, kind="normal")
    intensity_sigma_h                = Component(Signal, kind="normal")
    intensity_sigma_v                = Component(Signal, kind="normal")
    intensity_maximum                = Component(Signal, kind="normal")
    intensity_integral               = Component(Signal, kind="normal")
    phase_average_curvature_radius_h = Component(Signal, kind="normal")
    phase_average_curvature_radius_v = Component(Signal, kind="normal")
    beam                             = Component(Signal, kind="normal")

    def __init__(self,
                 prefix: str = "",
                 name: str = "beam_manager",
                 kind: Kind = Kind.normal,
                 beam_manager: IBeamManager = None,
                 verbose: bool = False):
        super(WavefrontAnalysisDetector, self).__init__(prefix=prefix,
                                                        name=name,
                                                        kind=kind,
                                                        beam_manager=beam_manager,
                                                        verbose=verbose)



# #########################################################################
# Copyright (c) 2023, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2023. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# #########################################################################

import os

try:
    from setuptools import find_packages, setup
except AttributeError:
    from setuptools import find_packages, setup

NAME = 'aps-beamline-driver'

VERSION = '1.0.0'
ISRELEASED = False

DESCRIPTION = 'APS Beamline Driver'
README_FILE = os.path.join(os.path.dirname(__file__), 'README.md')
LONG_DESCRIPTION = open(README_FILE).read()
AUTHOR = 'Luca Rebuffi'
AUTHOR_EMAIL = 'lrebuffi@anl.gov'
URL = 'https://github.com/APS-XSD-OPT-Group/Beamline-Driver'
DOWNLOAD_URL = 'https://github.com/APS-XSD-OPT-Group/Beamline-Driver'
MAINTAINER = 'XSD-OPT Group @ APS-ANL'
MAINTAINER_EMAIL = 'lrebuffi@anl.gov'
LICENSE = 'BSD-3'

KEYWORDS = ['dictionary',
    'glossary',
    'synchrotron'
    'simulation',
]

CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'Environment :: X11 Applications :: Qt',
    'Environment :: Console',
    'Environment :: Plugins',
    'Programming Language :: Python',
    'Operating System :: POSIX',
    'Operating System :: Microsoft :: Windows',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'Topic :: Scientific/Engineering :: Visualization',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Intended Audience :: Education',
    'Intended Audience :: Science/Research',
    'Intended Audience :: Developers',
]

INSTALL_REQUIRES = (
    'aps-common-libraries>=1.0.17',
    'cmasher',
)

PACKAGES = find_packages(exclude=('*.tests', '*.tests.*', 'tests.*', 'tests'))

PACKAGE_DATA = {
}

def setup_package():

    setup(
        name=NAME,
        version=VERSION,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type='text/markdown',
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        maintainer=MAINTAINER,
        maintainer_email=MAINTAINER_EMAIL,
        url=URL,
        download_url=DOWNLOAD_URL,
        license=LICENSE,
        keywords=KEYWORDS,
        classifiers=CLASSIFIERS,
        packages=PACKAGES,
        package_data=PACKAGE_DATA,
        install_requires=INSTALL_REQUIRES,
        zip_safe=False,
        include_package_data=True,
    )

if __name__ == '__main__':
    setup_package()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------- #
# Copyright (c) 2025, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2025. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# ----------------------------------------------------------------------- #
import os
import abc, decimal
from typing import List, Dict, Any
from enum import Enum
import json
import numpy as np

from aps.common.data_structures import DictionaryWrapper
from aps.beamline_driver.facade import MethodCallingType

class Movement(Enum):
    ABSOLUTE = 1
    RELATIVE = 2

class Units(Enum):
    MILLIRADIANS = 1
    DEGREES      = 2
    RADIANS      = 3
    MILLIMETERS  = 4
    MICRON       = 5
    METERS       = 6
    VOLTS        = 7
    MILLIVOLTS   = 8
    OTHER        = 9

    @property
    def label(self):
        if   self.value == 1: return "mrad"
        elif self.value == 2: return "deg"
        elif self.value == 3: return "rad"
        elif self.value == 4: return "mm"
        elif self.value == 5: return "\u03bcm"
        elif self.value == 6: return "m"
        elif self.value == 7: return "V"
        elif self.value == 8: return "mV"
        elif self.value == 9: return "Other"

    @classmethod
    def get_label(cls, value: int):
        if   value == 1: return "mrad"
        elif value == 2: return "deg"
        elif value == 3: return "rad"
        elif value == 4: return "mm"
        elif value == 5: return "\u03bcm"
        elif value == 6: return "m"
        elif value == 7: return "V"
        elif value == 8: return "mV"
        elif value == 9: return "Other"

    @classmethod
    def get_native_value(cls, value: int):
        if   value == 1: return Units.MILLIRADIANS
        elif value == 2: return Units.DEGREES
        elif value == 3: return Units.RADIANS
        elif value == 4: return Units.MILLIMETERS
        elif value == 5: return Units.MICRON
        elif value == 6: return Units.METERS
        elif value == 7: return Units.VOLTS
        elif value == 8: return Units.MILLIVOLTS
        elif value == 9: return Units.OTHER

    @classmethod
    def get_labels_list(cls):
        return ["mrad",
                "deg",
                "rad",
                "mm",
                "\u03bcm",
                "m",
                "V",
                "mV",
                "Other"]

class MotorType(Enum):
    ROTATIONAL    = 1
    TRANSLATIONAL = 2
    OTHER         = 3

    @classmethod
    def get_label(value: int):
        if   value == 1: return "Rotational"
        elif value == 2: return "Transational"
        elif value == 3: return "Other"

    @classmethod
    def get_native_value(cls, value: int):
        if   value == 1: return MotorType.ROTATIONAL
        elif value == 2: return MotorType.TRANSLATIONAL
        elif value == 3: return MotorType.OTHER

    @classmethod
    def get_labels_list(cls):
        return ["Rotational",
                "Transational",
                "Other"]


class MotorInfo(object):
    def __init__(self,
                 type :                 MotorType,
                 units :                Units,
                 resolution :           float,
                 repeatability:         float,
                 boundaries:            List[float],
                 shift_from_ideal:      float,
                 legacy_ids :           List[str]):
        self._type                  = type
        self._units                 = units
        self._resolution            = resolution
        self._repeatability         = repeatability
        self._boundaries            = boundaries
        self._shift_from_ideal      = shift_from_ideal
        self._legacy_ids            = legacy_ids

    @property
    def type(self) -> MotorType: return self._type
    @property
    def units(self) -> Units: return self._units
    @property
    def resolution(self) -> float: return self._resolution
    @property
    def repeatability(self) -> float: return self._repeatability
    @property
    def digits(self) -> int: return abs(decimal.Decimal(str(self._resolution)).as_tuple().exponent)
    @property
    def boundaries(self) -> List[float]: return self._boundaries
    @property
    def shift_from_ideal(self) -> float: return self._shift_from_ideal

    @property
    def legacy_ids(self) -> List[str]: return self._legacy_ids

    def apply_resolution(self, position):
        significant_digits = abs(decimal.Decimal(str(self._resolution)).as_tuple().exponent)
        return round(position, significant_digits)

    def __str__(self):
        return "type: " + str(self._type.name) + \
            ", units: " + str(self._units.label) + \
            ", resolution: " + str(self._resolution) + \
            ", repeatability: " + str(self._repeatability) + \
            ", boundaries: " + str(self._boundaries) + \
            ", shift_from_ideal: " + str(self._shift_from_ideal) + \
            ", legacy_ids: " + str(self._legacy_ids)

    def to_dictionary(self):
        return {"type":         self._type.value,
                "units":        self._units.value,
                "resolution":   self._resolution,
                "repeatability":   self._repeatability,
                "boundaries":   self._boundaries,
                "shift_from_ideal": self._shift_from_ideal,
                "legacy_ids":    self._legacy_ids}

    @classmethod
    def from_dictionary(self, dictionary: dict):
        return MotorInfo(type                  = MotorType(dictionary.get("type", 2)),
                         units                 = Units(dictionary.get("units", 9)),
                         resolution            = dictionary.get("resolution", 0.0),
                         repeatability         = dictionary.get("repeatability", 0.0),
                         boundaries            = dictionary.get("boundaries", [0.0, 0.0]),
                         shift_from_ideal      = dictionary.get("shift_from_ideal", 0.0),
                         legacy_ids            = dictionary.get("legacy_ids", ["", ""]))

class MotorsInfo(DictionaryWrapper):
    def __init__(self, **kwargs): super(MotorsInfo, self).__init__(**kwargs)

    def add_motor_info(self, motor_id : str, info : MotorInfo): self.set_parameter(motor_id, info)
    def get_motors_count(self) -> int: return self.get_parameters_number()
    def get_motor_info(self, motor_id : str) -> MotorInfo: return self.get_parameter(motor_id)
    def get_motor_ids(self) -> list: return self.get_parameter_names()
    def get_types(self, motor_ids: list=None) -> dict: return {motor_id: self.get_motor_info(motor_id).type for motor_id in (self.get_motor_ids() if motor_ids is None else motor_ids)}
    def get_units(self, motor_ids: list=None) -> dict: return {motor_id: self.get_motor_info(motor_id).units for motor_id in (self.get_motor_ids() if motor_ids is None else motor_ids)}
    def get_resolutions(self, motor_ids: list=None) -> dict: return {motor_id: self.get_motor_info(motor_id).resolution for motor_id in (self.get_motor_ids() if motor_ids is None else motor_ids)}
    def get_repeatabilities(self, motor_ids: list=None) -> dict: return {motor_id: self.get_motor_info(motor_id).repeatability for motor_id in (self.get_motor_ids() if motor_ids is None else motor_ids)}
    def get_boundaries(self, motor_ids: list=None) -> dict: return {motor_id: self.get_motor_info(motor_id).boundaries for motor_id in (self.get_motor_ids() if motor_ids is None else motor_ids)}
    def get_shifts_from_ideal(self, motor_ids: list=None) -> dict: return {motor_id: self.get_motor_info(motor_id).shift_from_ideal for motor_id in (self.get_motor_ids() if motor_ids is None else motor_ids)}

    def to_dictionary(self) -> dict: return {motor_id: self.get_motor_info(motor_id).to_dictionary() for motor_id in self.get_motor_ids()}

    @classmethod
    def from_dictionary(self, dictionary: dict):
        motors_info = MotorsInfo()
        for motor_id in dictionary.keys():
            motors_info.add_motor_info(motor_id=motor_id,
                                       info=MotorInfo.from_dictionary(dictionary.get(motor_id)))
        return motors_info

    @classmethod
    def from_json(cls, file_name: dict):
        with open(file_name, "r") as file: dictionary = json.load(file)
        return cls.from_dictionary(dictionary)

    def to_json(self, file_name: dict):
        with open(file_name, "w") as file: json.dump(self.to_dictionary(), file, indent=4)
        print(f"Motors Info saved to JSON file {file_name}")

    @classmethod
    def default_json_file_name(self, optical_system_id):
        return os.path.join(os.path.abspath(os.getcwd()), f"optical_system_{optical_system_id}_motion_manager.json")

#############################################################################################################
#
# LEGACY >> BEGIN
#
#############################################################################################################

class MotorInfoLegacy(object):
    def __init__(self,
                 type: MotorType,
                 units: Units,
                 resolution: float,
                 repeatability: float,
                 search_range: Dict[str, Dict[str, Dict[str, List[float]]]],
                 search_transfer_range: Dict[str, Dict[str, Dict[str, List[float]]]],
                 boundaries: Dict[str, List[float]],
                 initial_position: Dict[str, Dict[str, Dict[str, float]]],
                 shift_from_ideal: float,
                 legacy_ids: List[str],
                 optimized: Dict[str, Dict[str, Dict[str, bool]]]):
        self.__type                  = type
        self.__units                 = units
        self.__resolution            = resolution
        self.__repeatability         = repeatability
        self.__search_range          = search_range
        self.__search_transfer_range = search_transfer_range
        self.__boundaries            = boundaries
        self.__initial_position      = initial_position
        self.__shift_from_ideal      = shift_from_ideal
        self.__legacy_ids            = legacy_ids
        self.__optimized             = optimized

    def __get_value_from_key(self, dictionary, beam_manager_type, platform, beam_manager_id) -> Any:
        try:             return dictionary[beam_manager_type][platform].get(beam_manager_id, None)
        except KeyError: return None

    @property
    def type(self) -> MotorType: return self.__type
    @property
    def units(self) -> Units: return self.__units
    @property
    def resolution(self) -> float: return self.__resolution
    @property
    def repeatability(self) -> float: return self.__repeatability
    @property
    def digits(self) -> int: return abs(decimal.Decimal(str(self.__resolution)).as_tuple().exponent)
    @property
    def search_range(self) -> Dict[str, Dict[str, Dict[str, List[float]]]]: return self.__search_range
    @property
    def search_space(self) -> Dict[str, Dict[str, Dict[str, List[float]]]]:
        search_space = {}
        for beam_manager_type in self.__search_range.keys():
            search_space[beam_manager_type] = {}
            for platform in self.__search_range[beam_manager_type].keys():
                search_space[beam_manager_type][platform] = {}
                for beam_manager_id in self.__search_range[beam_manager_type][platform].keys():
                    search_range = self.__search_range[beam_manager_type][platform][beam_manager_id]
                    search_space[beam_manager_type][platform][beam_manager_id] = [round(item, self.digits) for item in np.arange(start=search_range[0],
                                                                                                                                 stop=search_range[1] + self.__resolution,
                                                                                                                                 step=self.__resolution)]
        return search_space

    def search_range_by_key(self, beam_manager_type, platform, beam_manager_id) -> List[float]:
        return self.__get_value_from_key(self.__search_range, beam_manager_type, platform, beam_manager_id)

    def search_space_by_key(self, beam_manager_type, platform, beam_manager_id) -> List[float]:
        try:
            search_range = self.__get_value_from_key(self.__search_range, beam_manager_type, platform, beam_manager_id)
            if not search_range is None:
                return [round(item, self.digits) for item in np.arange(start=search_range[0],
                                                                       stop=search_range[1] + self.__resolution,
                                                                       step=self.__resolution)]
            else: return None
        except KeyError:
            return None

    @property
    def search_transfer_range(self) -> Dict[str, Dict[str, Dict[str, List[float]]]]: return self.__search_transfer_range
    def search_transfer_range_by_key(self, beam_manager_type, platform, beam_manager_id) -> List[float]:
        return self.__get_value_from_key(self.__search_transfer_range, beam_manager_type, platform, beam_manager_id)

    @property
    def boundaries(self) -> Dict[str, List[float]]: return self.__boundaries
    def boundaries_by_key(self, platform) -> List[float]:
        return self.__boundaries.get(platform, None)

    @property
    def initial_position(self) -> Dict[str, Dict[str, Dict[str, float]]]: return self.__initial_position
    def initial_position_by_key(self, beam_manager_type, platform, beam_manager_id) -> float:
        return self.__get_value_from_key(self.__initial_position, beam_manager_type, platform, beam_manager_id)

    @property
    def shift_from_ideal(self) -> float: return self.__shift_from_ideal

    @property
    def legacy_ids(self) -> List[str]: return self.__legacy_ids
    @property
    def optimized(self) -> Dict[str, Dict[str, Dict[str, bool]]]: return self.__optimized
    def optimized_by_key(self, beam_manager_type, platform, beam_manager_id) -> bool:
        return self.__get_value_from_key(self.__optimized, beam_manager_type, platform, beam_manager_id)

    def apply_resolution(self, position):
        significant_digits = abs(decimal.Decimal(str(self.__resolution)).as_tuple().exponent)
        return round(position, significant_digits)

    def __str__(self):
        return "type: " + str(self.__type.name) + \
            ", units: " + str(self.__units.label) + \
            ", resolution: " + str(self.__resolution) + \
            ", repeatability: " + str(self.__repeatability) + \
            ", search_range: " + str(self.__search_range) + \
            ", search_transfer_range: " + str(self.__search_transfer_range) + \
            ", boundaries: " + str(self.__boundaries) + \
            ", initial_position: " + str(self.__initial_position) + \
            ", shift_from_ideal: " + str(self.__shift_from_ideal) + \
            ", legacy_ids: " + str(self.__legacy_ids)  + \
            ", optimized: " + str(self.__optimized)

    def to_dictionary(self):
        return {"type":         self.__type.value,
                "units":        self.__units.value,
                "resolution":   self.__resolution,
                "repeatability":   self.__repeatability,
                "search_range": self.__search_range,
                "search_transfer_range" : self.__search_transfer_range,
                "boundaries":   self.__boundaries,
                "initial_position":   self.__initial_position,
                "shift_from_ideal": self.__shift_from_ideal,
                "legacy_ids":    self.__legacy_ids,
                "optimized": self.__optimized}

    @classmethod
    def from_dictionary(self, dictionary: dict):
        return MotorInfoLegacy(type                  = MotorType(dictionary.get("type", 2)),
                               units                 = Units(dictionary.get("units", 9)),
                               resolution            = dictionary.get("resolution", 0.0),
                               repeatability         = dictionary.get("repeatability", 0.0),
                               search_range          = dictionary.get("search_range", {}),
                               search_transfer_range = dictionary.get("search_transfer_range", {}),
                               boundaries            = dictionary.get("boundaries", {}),
                               initial_position      = dictionary.get("initial_position", {}),
                               shift_from_ideal      = dictionary.get("shift_from_ideal", 0.0),
                               legacy_ids            = dictionary.get("legacy_ids", ["", ""]),
                               optimized             = dictionary.get("optimized", {}))

class MotorsInfoLegacy(DictionaryWrapper):
    def __init__(self, **kwargs): super(MotorsInfoLegacy, self).__init__(**kwargs)

    def add_motor_info(self, motor_id : str, info : MotorInfoLegacy): self.set_parameter(motor_id, info)
    def get_motors_count(self) -> int: return self.get_parameters_number()
    def get_motor_info(self, motor_id : str) -> MotorInfoLegacy: return self.get_parameter(motor_id)
    def get_motor_ids(self) -> list: return self.get_parameter_names()
    def get_types(self, motor_ids: list=None) -> dict: return {motor_id: self.get_motor_info(motor_id).type for motor_id in (self.get_motor_ids() if motor_ids is None else motor_ids)}
    def get_units(self, motor_ids: list=None) -> dict: return {motor_id: self.get_motor_info(motor_id).units for motor_id in (self.get_motor_ids() if motor_ids is None else motor_ids)}
    def get_resolutions(self, motor_ids: list=None) -> dict: return {motor_id: self.get_motor_info(motor_id).resolution for motor_id in (self.get_motor_ids() if motor_ids is None else motor_ids)}
    def get_repeatbilities(self, motor_ids: list=None) -> dict: return {motor_id: self.get_motor_info(motor_id).repeatability for motor_id in (self.get_motor_ids() if motor_ids is None else motor_ids)}
    def get_search_ranges(self, beam_manager_type, platform, beam_manager_id, motor_ids: list=None) -> dict:
        return {motor_id: self.get_motor_info(motor_id).search_range_by_key(beam_manager_type, platform, beam_manager_id) \
                for motor_id in (self.get_motor_ids() if motor_ids is None else motor_ids)}
    def get_search_spaces(self, beam_manager_type, platform, beam_manager_id, motor_ids: list=None) -> dict:
        return {motor_id: self.get_motor_info(motor_id).search_space_by_key(beam_manager_type, platform, beam_manager_id) \
                for motor_id in (self.get_motor_ids() if motor_ids is None else motor_ids)}
    def get_search_transfer_ranges(self, beam_manager_type, platform, beam_manager_id, motor_ids: list=None) -> dict:
        return {motor_id: self.get_motor_info(motor_id).search_transfer_range_by_key(beam_manager_type, platform, beam_manager_id) \
                for motor_id in (self.get_motor_ids() if motor_ids is None else motor_ids)}
    def get_boundaries(self, platform, motor_ids: list=None) -> dict:
        return {motor_id: self.get_motor_info(motor_id).boundaries_by_key(platform) \
                for motor_id in (self.get_motor_ids() if motor_ids is None else motor_ids)}
    def get_initial_positions(self, beam_manager_type, platform, beam_manager_id, motor_ids: list=None) -> dict:
        return {motor_id: self.get_motor_info(motor_id).initial_position_by_key(beam_manager_type, platform, beam_manager_id)
                for motor_id in (self.get_motor_ids() if motor_ids is None else motor_ids)}
    def get_shifts_from_ideal(self, motor_ids: list=None) -> dict: return {motor_id: self.get_motor_info(motor_id).shift_from_ideal for motor_id in (self.get_motor_ids() if motor_ids is None else motor_ids)}
    def get_optimized(self, beam_manager_type, platform, beam_manager_id, motor_ids: list=None) -> dict:
        return {motor_id: self.get_motor_info(motor_id).optimized_by_key(beam_manager_type, platform, beam_manager_id) \
                for motor_id in (self.get_motor_ids() if motor_ids is None else motor_ids)}
    def get_motor_ids_to_optimize(self, beam_manager_type, platform, beam_manager_id):
        motor_ids_to_optimize = []
        for motor_id in self.get_motor_ids():
            if self.get_motor_info(motor_id).optimized_by_key(beam_manager_type, platform, beam_manager_id): motor_ids_to_optimize.append(motor_id)
        return motor_ids_to_optimize

    def to_dictionary(self) -> dict: return {motor_id: self.get_motor_info(motor_id).to_dictionary() for motor_id in self.get_motor_ids()}
    @classmethod
    def from_dictionary(self, dictionary: dict):
        motors_info = MotorsInfoLegacy()
        for motor_id in dictionary.keys():
            motors_info.add_motor_info(motor_id=motor_id,
                                       info=MotorInfoLegacy.from_dictionary(dictionary.get(motor_id)))
        return motors_info

    @classmethod
    def is_legacy(cls, dictionary: dict) -> bool:
        for motor_id in dictionary.keys():
            try:
                _ = dictionary[motor_id]["search_range"]
                return True
            except KeyError:
                return False

#############################################################################################################
#
# LEGACY << END
#
#############################################################################################################

class IMotionManager():
    @abc.abstractmethod
    def method_calling_type(self) -> MethodCallingType: raise NotImplementedError
    @abc.abstractmethod
    def set_method_calling_type(self, value: MethodCallingType): raise NotImplementedError
    @abc.abstractmethod
    def move_motor(self, motor_id: str, position: float, movement : Movement, units : Units = None, *args, **kwargs): raise NotImplementedError
    @abc.abstractmethod
    def get_motor_position(self, motor_id: str, units : Units = None, *args, **kwargs) -> float: raise NotImplementedError
    @abc.abstractmethod
    def get_motors_info(self) -> MotorsInfo | MotorsInfoLegacy: raise NotImplementedError
    @abc.abstractmethod
    def move_motors(self, motor_positions : dict, movement : Movement, units: dict=None, *args, **kwargs): raise NotImplementedError
    @abc.abstractmethod
    def get_motor_positions(self, motor_ids: list=None, units: dict=None, *args, **kwargs) -> dict: raise NotImplementedError
    @abc.abstractmethod
    def to_json(self, file_name, *args, **kwargs): raise NotImplementedError
    @classmethod
    @abc.abstractmethod
    def from_json(cls, file_name, *args, **kwargs): raise NotImplementedError

# ------------------------------------------------------------------ #

def format_motor_positions(motion_manager : IMotionManager):
    text = "Motors Position:"

    info = motion_manager.get_motors_info()
    for id in info.get_motor_ids():
        text += "\n" + id + " = " + str(motion_manager.get_motor_position(motor_id=id)) + " " + str(info.get_motor_info(id).units.name)
    return text

def print_motor_positions(motion_manager : IMotionManager):
    print(format_motor_positions(motion_manager))


